@@ -7,6 +7,7 @@ use num_bigint::BigUint;
 use pyo3::{prelude::*, types::PyTuple};
 use rr_util::{
     atr,
+    compact_data::U8Set,
     rearrange_spec::RInnerInts,
     tensor_util::{IndexError, IndexTensor, TensorIndexSync},
     tu8v,
@@ -60,8 +61,7 @@ circuit_node_extra_impl!(Einsum);
 
 impl CircuitNodeComputeInfoImpl for Einsum {
     fn compute_shape(&self) -> Shape {
-        let map = self.shape_map().unwrap();
-        self.out_axes.iter().map(|a| *map.get(a).unwrap()).collect()
+        self.info().shape.clone()
     }
 }
 impl CircuitNodeHashWithChildren for Einsum {
@@ -197,7 +197,7 @@ impl Einsum {
     #[apply(new_rc_unwrap)]
     pub fn try_new(args: EinsumArgs, out_axes: EinsumAxes, name: Option<String>) -> Result<Self> {
         // TODO: reduce axes nums!
-        let out = Self {
+        let mut out = Self {
             args,
             out_axes,
             name,
@@ -206,18 +206,19 @@ impl Einsum {
 
         let map = out.shape_map()?; // catch errors
 
-        for a in &out.out_axes {
-            if !map.contains_key(a) {
-                return Err(ConstructError::EinsumOutputNotSubset {
+        out.info.shape = out
+            .out_axes
+            .iter()
+            .map(|a| map.get(a).cloned())
+            .collect::<Option<Shape>>()
+            .ok_or_else(|| {
+                anyhow::Error::from(ConstructError::EinsumOutputNotSubset {
                     circuit_name: out.name.clone(),
                     all_input_axes: map.into_keys().collect(),
                     output_axes: out.out_axes.clone(),
-                }
-                .into());
-            }
-        }
+                })
+            })?;
 
-        // could reuse hashmap if helped with perf
         out.initial_init_info()
     }
 
@@ -487,7 +488,7 @@ impl Einsum {
         .unwrap()
     }
 
-    pub fn reduced_axes(&self) -> HashSet<u8> {
+    pub fn reduced_axes(&self) -> U8Set {
         self.input_axes()
             .flatten()
             .filter(|i| !self.out_axes.contains(i))
@@ -1262,6 +1263,14 @@ impl Index {
             .collect()];
         result
     }
+
+    /// this is the same as "is view (not copy) on child"
+    pub fn has_tensor_axes(&self) -> bool {
+        self.index
+            .0
+            .iter()
+            .any(|x| matches!(x, TensorAxisIndex::Tensor(_)))
+    }
 }
 
 #[pyclass(extends=PyCircuitBase)]
