@@ -953,15 +953,17 @@ impl<F: ReplaceForExpand> ReplaceExpander<F> {
             return Ok(replaced.clone());
         }
 
+        let per_child_extra = extra_replacements.per_child(&circuit);
+
         // TODO: optimize this as needed
         let new_children = circuit
             .children()
-            .zip(extra_replacements.per_child(&circuit))
+            .zip(per_child_extra.clone())
             .map(|(c, rep)| self_.replace_expand_with_map(c, &rep))
             .collect::<Result<_>>()?;
 
-        expand_node(circuit, &new_children, |c, rep, _child_idx| {
-            self_.replace_expand_with_map(c, &extra_replacements.extend_into(rep))
+        expand_node(circuit, &new_children, |c, rep, child_idx| {
+            self_.replace_expand_with_map(c, &per_child_extra[child_idx].extend_into(rep))
         })
     }
 }
