@@ -32,6 +32,7 @@ use pyo3::{
 };
 use rr_util::{
     cached_lambda,
+    compact_data::U8Set,
     eq_by_big_hash::EqByBigHash,
     py_types::{
         reduction_to_ints, use_rust_comp, PyCallable, PyOpAtAxes, PyShape, Tensor, PY_CIRCUIT_ITEMS,
@@ -1405,6 +1406,9 @@ pub enum ConstructError {
     #[error("Conv input wrong shape ({e_name})")]
     ConvFilterWrongShape {},
 
+    #[error("Array has an axis length with a prime factor above 10,000, which are reserved for symbolic shapes, number {number} symbolic meaning {sym} ({e_name})")]
+    ArrayHasReservedSymbolicShape { number: usize, sym: String },
+
     #[error("Conv kernel shape must be all odd, got ({e_name})")]
     ConvFilterMustBeOddLength { shape: Shape },
 
@@ -1469,7 +1473,7 @@ pub enum ConstructError {
     #[error("circuit_name={circuit_name:?} all_input_axes={all_input_axes:?} output_axes={output_axes:?} ({e_name})")]
     EinsumOutputNotSubset {
         circuit_name: Option<String>,
-        all_input_axes: HashSet<u8>,
+        all_input_axes: U8Set,
         output_axes: EinsumAxes,
     },
 
