@@ -14,6 +14,7 @@ use itertools::Itertools;
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
+    compact_data::U8Set,
     filter_by_variant,
     opt_einsum::optimize_einsum_spec_cached,
     py_types::PY_CIRCUIT_ITEMS,
@@ -733,7 +734,7 @@ pub fn einsum_pull_removable_axes(einsum: &Einsum) -> Option<CircuitRc> {
     if !did_anything {
         return None;
     }
-    let ints_in_new_args: HashSet<u8> = new_args
+    let ints_in_new_args: U8Set = new_args
         .iter()
         .flat_map(|(_node, ints)| ints.as_slice())
         .cloned()
@@ -743,9 +744,9 @@ pub fn einsum_pull_removable_axes(einsum: &Einsum) -> Option<CircuitRc> {
     let shape_map = einsum.shape_map().unwrap();
     let reduced_axes = einsum.reduced_axes();
     let scalar_mul: f64 = reduced_axes
-        .iter()
-        .filter(|i| !ints_in_new_args.contains(i))
-        .map(|i| shape_map[i] as f64)
+        .into_iter()
+        .filter(|i| !ints_in_new_args.contains(*i))
+        .map(|i| shape_map[&i] as f64)
         .product();
     if scalar_mul != 1.0 {
         new_args.push((Scalar::new(scalar_mul, sv![], None).rc(), tu8v![]));
@@ -754,7 +755,7 @@ pub fn einsum_pull_removable_axes(einsum: &Einsum) -> Option<CircuitRc> {
     let new_out_ints: EinsumAxes = einsum
         .out_axes
         .iter()
-        .filter(|i| ints_in_new_args.contains(i))
+        .filter(|i| ints_in_new_args.contains(**i))
         .copied()
         .collect();
     let new_einsum = Einsum::nrc(new_args, new_out_ints.clone(), einsum.name_cloned());
@@ -1077,8 +1078,7 @@ pub fn einsum_elim_zero(einsum: &Einsum) -> Option<Scalar> {
     }
 }
 
-#[pyfunction]
-pub fn einsum_merge_scalars(einsum: &Einsum) -> Option<Einsum> {
+pub fn einsum_merge_scalars_impl(einsum: &Einsum, only_one: bool) -> Option<Einsum> {
     let mut num_scalars_found = 0;
     let mut scalar_mul: f64 = 1.0;
     let mut new_args: Vec<(CircuitRc, EinsumAxes)> = einsum
@@ -1086,7 +1086,7 @@ pub fn einsum_merge_scalars(einsum: &Einsum) -> Option<Einsum> {
         .iter()
         .filter(|(node, _ints)| match &***node {
             Circuit::Scalar(scalar) => {
-                if scalar.info().shape.is_empty() {
+                if scalar.info().shape.is_empty() && (!only_one || scalar.value == 1.) {
                     num_scalars_found += 1;
                     scalar_mul *= scalar.value;
                     false
@@ -1098,7 +1098,7 @@ pub fn einsum_merge_scalars(einsum: &Einsum) -> Option<Einsum> {
         })
         .cloned()
         .collect();
-    if num_scalars_found <= 1 {
+    if num_scalars_found == 0 || (num_scalars_found == 1 && scalar_mul != 1.) {
         return None;
     }
     if scalar_mul != 1.0 {
@@ -1107,6 +1107,16 @@ pub fn einsum_merge_scalars(einsum: &Einsum) -> Option<Einsum> {
     Some(Einsum::new(new_args, einsum.out_axes.clone(), None))
 }
 
+#[pyfunction]
+pub fn einsum_merge_scalars(einsum: &Einsum) -> Option<Einsum> {
+    einsum_merge_scalars_impl(einsum, false)
+}
+
+#[pyfunction]
+pub fn einsum_remove_one(einsum: &Einsum) -> Option<Einsum> {
+    einsum_merge_scalars_impl(einsum, true)
+}
+
 /// We want to turn einsum ab->b, where a is concat, to be Add [a1b->b, a2b->b...]
 /// This allows us to not realize all of ab at once to save memory.
 #[pyfunction]
