@@ -23,18 +23,24 @@ use crate::{
     circuit_optimizer::{OptimizationContext, OptimizationSettings},
 };
 
-#[pyfunction(keep_non_axis_leaves = "false")]
+#[pyfunction(keep_non_axis_leaves = "false", allow_partial_batch = "true")]
 pub fn batch_to_concat(
     circuit: CircuitRc,
     axis: usize,
-    num_batches: usize,
+    batch_size: usize,
     keep_non_axis_leaves: bool, // THIS IS A HACK
+    allow_partial_batch: bool,
 ) -> Result<CircuitRc> {
+    if axis >= circuit.info().shape.len() {
+        bail!("batch_to_concat on nonexistent axis");
+    }
     let l = circuit.info().shape[axis];
-    if l % num_batches != 0 {
-        bail!(BatchError::NumberDoesntDivide { l, num_batches });
+    if batch_size == 0 {
+        bail!(BatchError::BatchSizeIsZero {});
+    }
+    if l % batch_size != 0 && !allow_partial_batch {
+        bail!(BatchError::PartialBatchNotAllowed {});
     }
-    let batch_size = l / num_batches;
     let circuit = deep_map_unwrap(circuit, |x| {
         x.as_index()
             .map(|x| Index::slice_edges_to_none(x).rc())
@@ -47,15 +53,7 @@ pub fn batch_to_concat(
     } else {
         leaves_non_axis
     };
-    // println!("leaves axis");
-    // for (k, v) in &leaves_axis {
-    //     println!("{}", v);
-    //     k.print();
-    // }
-    // println!("leaves non axis");
-    // for k in &leaves_non_axis {
-    //     k.print();
-    // }
+
     let input_specs_axis = leaves_axis
         .iter()
         .map(|(sub, i)| {
@@ -87,10 +85,6 @@ pub fn batch_to_concat(
             )
         }))
         .collect();
-    // println!("input_specs");
-    // for is in &input_specs {
-    //     is.0.print();
-    // }
     if input_specs.is_empty() {
         bail!(BatchError::AxisOriginatesTooHigh {});
     }
@@ -102,66 +96,74 @@ pub fn batch_to_concat(
         false,
     )
     .unwrap();
-    // println!("module spec");
-    // module_spec.spec_circuit.print();
-    let module_spec = module_spec.resize(
+
+    let mk_module_spec = |axis_size: usize| -> Result<ModuleSpec> {
+        module_spec.resize(
+            module_spec
+                .arg_specs
+                .iter()
+                .map(|inp_spec| {
+                    if let Some((c, _is)) = input_specs_axis
+                        .iter()
+                        .find(|(_c, spec)| spec.symbol.uuid == inp_spec.symbol.uuid)
+                    {
+                        let mut result = c.info().shape.clone();
+                        result[leaves_axis[c]] = axis_size;
+                        return result;
+                    }
+                    inp_spec.symbol.info().shape.clone()
+                })
+                .collect(),
+        )
+    };
+    let mk_args = |index: TensorAxisIndex| -> Vec<CircuitRc> {
         module_spec
             .arg_specs
             .iter()
             .map(|inp_spec| {
                 if let Some((c, _is)) = input_specs_axis
                     .iter()
-                    .find(|(_c, spec)| spec.symbol.uuid == inp_spec.symbol.uuid)
+                    .find(|z| z.1.symbol.uuid == inp_spec.symbol.uuid)
                 {
-                    let mut result = c.info().shape.clone();
-                    result[leaves_axis[c]] = batch_size;
-                    return result;
+                    return Index::nrc(
+                        c.clone(),
+                        TensorIndex::new_single(index.clone(), leaves_axis[c], c.info().rank()),
+                        None,
+                    );
                 }
-                inp_spec.symbol.info().shape.clone()
+                input_specs
+                    .iter()
+                    .find(|(_c, s)| s.symbol.uuid == inp_spec.symbol.uuid)
+                    .unwrap()
+                    .0
+                    .clone()
             })
-            .collect(),
-    )?;
-    // println!("module spec after resize");
-    // module_spec.spec_circuit.print();
-    let concattands: Vec<CircuitRc> = (0..num_batches)
+            .collect()
+    };
+    let module_spec_main = mk_module_spec(batch_size)?;
+    let num_even_batches = l / batch_size;
+    let mut concattands: Vec<CircuitRc> = (0..num_even_batches)
         .map(|i| {
-            let new_nodes: Vec<CircuitRc> = module_spec
-                .arg_specs
-                .iter()
-                .map(|inp_spec| {
-                    if let Some((c, _is)) = input_specs_axis
-                        .iter()
-                        .find(|z| z.1.symbol.uuid == inp_spec.symbol.uuid)
-                    {
-                        return Index::nrc(
-                            c.clone(),
-                            TensorIndex::new_single(
-                                TensorAxisIndex::new_plain_slice(
-                                    i * batch_size,
-                                    (i + 1) * batch_size,
-                                ),
-                                leaves_axis[c],
-                                c.info().rank(),
-                            ),
-                            None,
-                        );
-                    }
-                    input_specs
-                        .iter()
-                        .find(|(_c, s)| s.symbol.uuid == inp_spec.symbol.uuid)
-                        .unwrap()
-                        .0
-                        .clone()
-                })
-                .collect();
-
-            // println!("module args");
-            // for c in new_nodes.iter() {
-            //     c.print();
-            // }
-            Module::nrc(new_nodes, module_spec.clone(), None)
+            Module::nrc(
+                mk_args(TensorAxisIndex::new_plain_slice(
+                    i * batch_size,
+                    (i + 1) * batch_size,
+                )),
+                module_spec_main.clone(),
+                None,
+            )
         })
         .collect();
+    if l % batch_size != 0 {
+        concattands.push(Module::nrc(
+            mk_args(TensorAxisIndex::new_plain_slice(
+                num_even_batches * batch_size,
+                l,
+            )),
+            mk_module_spec(l % batch_size)?,
+            None,
+        ));
+    }
     let result = Concat::nrc(concattands, axis, None);
     if result.info().shape != circuit.info().shape {
         println!(
@@ -183,14 +185,17 @@ pub fn batch_to_concat(
 #[base_exception(PyValueError)]
 #[derive(Error, Debug, Clone)]
 pub enum BatchError {
-    #[error("num_batches {num_batches} doesn't divide length {l} ({e_name})")]
-    NumberDoesntDivide { l: usize, num_batches: usize },
-
     #[error("Would need to batch multiple axes, only supports one ({e_name})")]
     RequiresMultipleAxes {},
 
     #[error("Batching axis originates too high ({e_name})")]
     AxisOriginatesTooHigh {},
+
+    #[error("batch_size must be greater than 0 ({e_name})")]
+    BatchSizeIsZero {},
+
+    #[error("axis not evenly dividable by batch_size and allow_partial_batch is False ({e_name})")]
+    PartialBatchNotAllowed {},
 }
 
 #[pyfunction]
@@ -265,9 +270,9 @@ pub fn batch_einsum(einsum: &Einsum, context: &mut OptimizationContext) -> Resul
         .map(|x| x.info().numel_usize())
         .max()
         .unwrap();
-    let mut num_batches = ((max_interm_mem as f64) / (max_single_mem as f64)) as usize;
-    while batch_int_len % num_batches != 0 {
-        num_batches += 1;
+    let mut batch_size = ((max_interm_mem as f64) / (max_single_mem as f64)) as usize;
+    while batch_int_len % batch_size != 0 {
+        batch_size -= 1;
     }
 
     let mut batchheads: Vec<CircuitRc> = vec![];
@@ -286,7 +291,8 @@ pub fn batch_einsum(einsum: &Einsum, context: &mut OptimizationContext) -> Resul
             batch_to_concat(
                 c.clone(),
                 axis_of_name(c, &batch_str).unwrap(),
-                num_batches,
+                batch_size,
+                false,
                 false,
             )
             .unwrap()
