@@ -3,7 +3,9 @@ use std::io::prelude::*;
 use anyhow::{bail, Context, Result};
 use circuit_base::{
     flat_concat,
-    module::{any_children_with_symbolic_sizes, conform_all_modules},
+    module::{
+        any_children_with_symbolic_sizes, conform_all_modules, inline_single_callsite_modules,
+    },
     print::{print_circuit_stats, PrintOptions},
     CircuitNode, CircuitRc, ModuleSpec,
 };
@@ -23,7 +25,7 @@ use rustc_hash::FxHashMap as HashMap;
 use crate::{
     canonicalize::deep_canonicalize,
     compiler_heuristics::deep_maybe_distribute,
-    compiler_strip::{remove_autotags, strip_names},
+    compiler_strip::strip_names_and_tags,
     deep_rewrite::{
         compiler_simp, deep_heuristic_nest_adds, deep_optimize_einsums, deep_push_down_index_raw,
     },
@@ -237,9 +239,9 @@ pub fn optimize_circuit(
     context: &mut OptimizationContext,
 ) -> Result<CircuitRc> {
     let print_timings = context.settings.verbose >= 2;
-    let circuit = timed!(strip_names(circuit, context, true), 10, print_timings);
-    let circuit = remove_autotags(circuit);
+    let circuit = timed!(strip_names_and_tags(circuit, context), 10, print_timings);
     let circuit = conform_all_modules(circuit);
+    let circuit = inline_single_callsite_modules(circuit);
     if is_python_running() && any_children_with_symbolic_sizes(circuit.clone()) {
         Python::with_gil(|py| {
             pyo3::PyErr::warn(
