@@ -9,6 +9,7 @@ use get_update_node::{
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
+    compact_data::U8Set,
     python_error_exception,
     rearrange_spec::{check_permutation, PermError},
     union_find::UnionFind,
@@ -1153,7 +1154,7 @@ pub fn nest_flat_einsum_strict_rec(
     name: Option<String>,
 ) -> Result<Einsum> {
     // TODO: avoid quadratic running time...
-    let all_ints: Vec<HashSet<AxisInt>> = specs
+    let all_ints: Vec<U8Set> = specs
         .iter()
         .map(|spec| {
             spec.idxs
@@ -1174,34 +1175,33 @@ pub fn nest_flat_einsum_strict_rec(
                 NestIdxsItem::Single(v) => flat_einsum[v].clone(),
                 NestIdxsItem::Many(sub_specs) => {
                     let my_ints = &all_ints[i];
-                    let all_other_ints: HashSet<AxisInt> = all_ints
+                    let all_other_ints: U8Set = all_ints
                         .iter()
                         .enumerate()
                         .filter(|(j, _)| j != &i)
-                        .flat_map(|(_, ints)| ints)
-                        .cloned()
+                        .flat_map(|(_, ints)| *ints)
                         .collect();
-                    let all_other_ints: HashSet<_> =
-                        all_other_ints.union(&out_axes_set).cloned().collect();
-                    let out_ints = all_other_ints.intersection(my_ints).cloned();
+                    let all_other_ints: U8Set = all_other_ints.union(&out_axes_set);
+                    let out_ints = all_other_ints.intersection(my_ints);
 
                     let mut sub_out_axes = if let Some(sub_out_axes) = spec.info.out_axes {
-                        let out_ints_actual = sub_out_axes.iter().cloned().collect::<HashSet<_>>();
+                        let out_ints_actual: U8Set =
+                            sub_out_axes.iter().cloned().collect::<U8Set>();
                         assert!(out_ints_actual.is_subset(my_ints));
-                        let minimal_out_num_set: HashSet<u8> = out_ints.collect();
+                        let minimal_out_num_set: U8Set = out_ints;
                         assert!(out_ints_actual.is_superset(&minimal_out_num_set));
                         if spec.info.shrink_out_axes {
                             // keep ordering for user simplicity
                             sub_out_axes
                                 .iter()
-                                .filter(|x| minimal_out_num_set.contains(*x))
+                                .filter(|x| minimal_out_num_set.contains(**x))
                                 .cloned()
                                 .collect()
                         } else {
                             sub_out_axes
                         }
                     } else {
-                        let mut sub_out_axes: EinsumAxes = out_ints.collect();
+                        let mut sub_out_axes: EinsumAxes = out_ints.into_iter().collect();
                         sub_out_axes.as_mut_slice().sort();
                         sub_out_axes
                     };
