@@ -17,7 +17,7 @@ use thiserror::Error;
 use uuid::uuid;
 
 use crate::{
-    compact_data::TinyVecU8,
+    compact_data::{TinyVecU8, U8Set},
     make_single_many,
     py_types::{einops_repeat, PyOpAtAxes, Tensor, PY_CIRCUIT_ITEMS},
     python_error_exception, sv,
@@ -657,7 +657,7 @@ impl RearrangeSpec {
         format!("{:?}", self)
     }
     pub fn shapes(&self) -> Result<(Shape, Shape)> {
-        let used_ints: HashSet<u8> = self
+        let used_ints: U8Set = self
             .input_ints
             .iter()
             .chain(&self.output_ints)
@@ -668,7 +668,7 @@ impl RearrangeSpec {
             .int_sizes
             .iter()
             .enumerate()
-            .any(|(i, x)| x.is_none() && used_ints.contains(&(i as u8)))
+            .any(|(i, x)| x.is_none() && used_ints.contains(i as u8))
         {
             bail!(RearrangeSpecError::HasWildcardSizes { spec: self.clone() });
         }
